from __future__ import annotations

import json
import logging
import os

import boto3
import structlog


def pretty_log(logger, method_name, event_dict):
    return_dict = {}

    # Uncomment this if we want to default to single line logging. Can get hard to read with long resources though.
    # if not any(bool(isinstance(value, dict) or isinstance(value, list)) for value in event_dict.values()):
    #     return event_dict

    for key, value in event_dict.items():
        if key != "event":
            key = f"\n  {key}"

        if isinstance(value, dict) or isinstance(value, list):
            value = "\n  ".join(str(json.dumps(value, indent=2)).split("\n"))

        return_dict[key] = value

    return return_dict


def configure_logger(logger_name, log_level):
    if not structlog.is_configured():
        structlog.configure(
            processors=[
                pretty_log,
                structlog.processors.add_log_level,
                structlog.processors.StackInfoRenderer(),
                structlog.dev.set_exc_info,
                structlog.processors.TimeStamper("%Y/%m/%d %H:%M:%S", utc=False),
                structlog.dev.ConsoleRenderer(),
            ],
            wrapper_class=structlog.make_filtering_bound_logger(
                logging.getLevelName(log_level)
            ),
            context_class=dict,
            logger_factory=structlog.PrintLoggerFactory(),
            cache_logger_on_first_use=False,
        )
    log = structlog.get_logger(logger_name)

    if log_level == "DEBUG":
        boto3.set_stream_logger("", "DEBUG")
    else:
        default_logging_levels = {
            "boto3": "CRITICAL",
            "boto": "CRITICAL",
            "botocore": "CRITICAL",
        }
        for logger, level in default_logging_levels.items():
            logging.getLogger(logger).setLevel(level)

    return log


log = configure_logger("iambic", os.getenv("LOG_LEVEL", "INFO"))



from __future__ import annotations

import asyncio
import json
from itertools import chain
from typing import Union

from deepdiff import DeepDiff

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import aio_wrapper, plugin_apply_wrapper
from iambic.plugins.v0_1_0.aws.models import AWSAccount
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call, paginated_search


async def get_user_inline_policy_names(user_name: str, iam_client):
    return await paginated_search(
        iam_client.list_user_policies, "PolicyNames", UserName=user_name
    )


async def list_users(iam_client):
    # user_details_list is missing MaxSessionDuration, see https://docs.aws.amazon.com/IAM/latest/APIReference/API_RoleDetail.html
    return await paginated_search(
        iam_client.get_account_authorization_details, "UserDetailList", Filter=["User"]
    )


async def list_user_tags(user_name: str, iam_client):
    return await paginated_search(iam_client.list_user_tags, "Tags", UserName=user_name)


async def get_user_policy(user_name: str, policy_name: str, iam_client):
    return await boto_crud_call(
        iam_client.get_user_policy, UserName=user_name, PolicyName=policy_name
    )


async def get_user_groups(
    user_name: str, iam_client, as_dict: bool = True
) -> Union[list, dict]:
    groups = await paginated_search(
        iam_client.list_groups_for_user, "Groups", UserName=user_name
    )
    if as_dict:
        return {group["GroupName"]: group for group in groups}
    else:
        return [{"GroupName": group["GroupName"]} for group in groups]


async def get_user_inline_policies(
    user_name: str, iam_client, as_dict: bool = True
) -> Union[list, dict]:
    policy_names = await get_user_inline_policy_names(user_name, iam_client)
    policies = await asyncio.gather(
        *[
            get_user_policy(user_name, policy_name, iam_client)
            for policy_name in policy_names
        ]
    )
    if as_dict:
        return {policy["PolicyName"]: policy["PolicyDocument"] for policy in policies}
    else:
        return [
            {"PolicyName": policy["PolicyName"], **policy["PolicyDocument"]}
            for policy in policies
        ]


async def get_user_managed_policies(user_name: str, iam_client) -> list[dict[str, str]]:
    marker: dict[str, str] = {}
    policies = []

    while True:
        # TODO: Figure out why attached managed policies are not shown
        response = await boto_crud_call(
            iam_client.list_attached_user_policies, UserName=user_name, **marker
        )
        policies.extend(response["AttachedPolicies"])

        if response["IsTruncated"]:
            marker["Marker"] = response["Marker"]
        else:
            break

    return [{"PolicyArn": policy["PolicyArn"]} for policy in policies]


async def get_user(user_name: str, iam_client, include_policies: bool = True) -> dict:
    try:
        current_user = (await boto_crud_call(iam_client.get_user, UserName=user_name))[
            "User"
        ]
        if include_policies:
            current_user["ManagedPolicies"] = await get_user_managed_policies(
                user_name, iam_client
            )
            current_user["InlinePolicies"] = await get_user_inline_policies(
                user_name, iam_client, as_dict=False
            )
            current_user["Groups"] = await get_user_groups(
                user_name, iam_client, as_dict=False
            )
    except iam_client.exceptions.NoSuchEntityException:
        current_user = {}

    return current_user


async def get_user_across_accounts(
    aws_accounts: list[AWSAccount], user_name: str, include_policies: bool = True
) -> dict:
    async def get_user_for_account(aws_account: AWSAccount):
        iam_client = await aws_account.get_boto3_client("iam")
        return {
            aws_account.account_id: await get_user(
                user_name, iam_client, include_policies
            )
        }

    account_on_users = await asyncio.gather(
        *[get_user_for_account(aws_account) for aws_account in aws_accounts]
    )
    return {
        account_id: user
        for resp in account_on_users
        for account_id, user in resp.items()
    }


async def apply_user_tags(
    user_name: str,
    iam_client,
    template_tags: list[dict],
    existing_tags: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    existing_tag_map = {tag["Key"]: tag.get("Value") for tag in existing_tags}
    template_tag_map = {tag["Key"]: tag.get("Value") for tag in template_tags}
    tags_to_apply = [
        tag
        for tag in template_tags
        if tag.get("Value") != existing_tag_map.get(tag["Key"])
    ]
    tasks = []
    response = []

    if tags_to_remove := [
        tag["Key"] for tag in existing_tags if not template_tag_map.get(tag["Key"])
    ]:
        log_str = "Stale tags discovered."

        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                resource_type="aws:iam:user",
                resource_id=user_name,
                attribute="tags",
                change_summary={"TagKeys": tags_to_remove},
            )
        ]

        response.extend(proposed_changes)

        if ctx.execute:
            log_str = f"{log_str} Removing tags..."

            apply_awaitable = boto_crud_call(
                iam_client.untag_user,
                UserName=user_name,
                TagKeys=tags_to_remove,
            )
            tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))

        log.debug(log_str, tags=tags_to_remove, **log_params)

    if tags_to_apply:
        log_str = "New tags discovered in AWS."

        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_type="aws:iam:user",
                resource_id=user_name,
                attribute="tags",
                new_value=tag,
                current_value=existing_tag_map.get(tag["Key"]),
            )
            for tag in tags_to_apply
        ]
        response.extend(proposed_changes)
        if ctx.execute:
            log_str = f"{log_str} Adding tags..."
            apply_awaitable = boto_crud_call(
                iam_client.tag_user, UserName=user_name, Tags=tags_to_apply
            )
            tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))

        log.debug(log_str, tags=tags_to_apply, **log_params)

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def apply_user_permission_boundary(
    user_name,
    iam_client,
    template_permission_boundary: dict,
    existing_permission_boundary: dict,
    log_params: dict,
) -> list[ProposedChange]:
    tasks = []
    response = []
    template_boundary_policy_arn = template_permission_boundary.get(
        "PolicyArn"
    )  # from serializing iambic template
    existing_boundary_policy_arn = existing_permission_boundary.get(
        "PermissionsBoundaryArn"
    )  # from boto response

    if template_boundary_policy_arn and (
        existing_boundary_policy_arn != template_boundary_policy_arn
    ):
        log_str = "New or updated permission boundary discovered."

        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_type="aws:policy_document",
                resource_id=template_boundary_policy_arn,
                attribute="permission_boundary",
            )
        ]
        response.extend(proposed_changes)

        if ctx.execute:
            log_str = f"{log_str} Attaching permission boundary..."

            tasks = [
                plugin_apply_wrapper(
                    boto_crud_call(
                        iam_client.put_user_permissions_boundary,
                        UserName=user_name,
                        PermissionsBoundary=template_boundary_policy_arn,
                    ),
                    proposed_changes,
                )
            ]

        log.debug(
            log_str, permission_boundary=template_boundary_policy_arn, **log_params
        )

    # Detach permission boundary not in template
    if template_boundary_policy_arn is None and (
        existing_boundary_policy_arn != template_boundary_policy_arn
    ):
        log_str = "Stale permission boundary discovered."

        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                resource_type="aws:policy_document",
                resource_id=existing_boundary_policy_arn,
                attribute="permission_boundary",
            )
        ]
        response.extend(proposed_changes)

        if ctx.execute:
            log_str = f"{log_str} Detaching permission boundary..."

            tasks.extend(
                [
                    plugin_apply_wrapper(
                        boto_crud_call(
                            iam_client.delete_user_permissions_boundary,
                            UserName=user_name,
                        ),
                        proposed_changes,
                    )
                ]
            )

        log.debug(
            log_str, permission_boundary=existing_boundary_policy_arn, **log_params
        )

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def apply_user_managed_policies(
    user_name,
    iam_client,
    template_policies: list[dict],
    existing_policies: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    tasks = []
    response = []
    template_policies = [policy["PolicyArn"] for policy in template_policies]
    existing_managed_policies = [policy["PolicyArn"] for policy in existing_policies]

    # Create new managed policies
    new_managed_policies = [
        policy_arn
        for policy_arn in template_policies
        if policy_arn not in existing_managed_policies
    ]
    if new_managed_policies:
        log_str = "New managed policies discovered."

        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_type="aws:policy_document",
                resource_id=policy_arn,
                attribute="managed_policies",
            )
            for policy_arn in new_managed_policies
        ]
        response.extend(proposed_changes)

        if ctx.execute:
            log_str = f"{log_str} Attaching managed policies..."

            tasks = [
                plugin_apply_wrapper(
                    boto_crud_call(
                        iam_client.attach_user_policy,
                        UserName=user_name,
                        PolicyArn=policy_arn,
                    ),
                    [
                        ProposedChange(
                            change_type=ProposedChangeType.ATTACH,
                            resource_type="aws:policy_document",
                            resource_id=policy_arn,
                            attribute="managed_policies",
                        )
                    ],
                )
                for policy_arn in new_managed_policies
            ]

        log.debug(log_str, managed_policies=new_managed_policies, **log_params)

    # Delete existing managed policies not in template
    existing_managed_policies = [
        policy_arn
        for policy_arn in existing_managed_policies
        if policy_arn not in template_policies
    ]
    if existing_managed_policies:
        log_str = "Stale managed policies discovered."

        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                resource_type="aws:policy_document",
                resource_id=policy_arn,
                attribute="managed_policies",
            )
            for policy_arn in existing_managed_policies
        ]
        response.extend(proposed_changes)

        if ctx.execute:
            log_str = f"{log_str} Detaching managed policies..."

            tasks.extend(
                [
                    plugin_apply_wrapper(
                        boto_crud_call(
                            iam_client.detach_user_policy,
                            UserName=user_name,
                            PolicyArn=policy_arn,
                        ),
                        [
                            ProposedChange(
                                change_type=ProposedChangeType.DETACH,
                                resource_type="aws:policy_document",
                                resource_id=policy_arn,
                                attribute="managed_policies",
                            )
                        ],
                    )
                    for policy_arn in existing_managed_policies
                ]
            )

        log.debug(log_str, managed_policies=existing_managed_policies, **log_params)

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def apply_user_inline_policies(
    user_name,
    iam_client,
    template_policies: list[dict],
    existing_policies: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    apply_tasks = []
    delete_tasks = []
    plan_response = []
    template_policy_map = {
        policy["PolicyName"]: {k: v for k, v in policy.items() if k != "PolicyName"}
        for policy in template_policies
    }
    existing_policy_map = {
        policy["PolicyName"]: {k: v for k, v in policy.items() if k != "PolicyName"}
        for policy in existing_policies
    }

    for policy_name in existing_policy_map.keys():
        if not template_policy_map.get(policy_name):
            log_str = "Stale inline policies discovered."
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.DELETE,
                    resource_type="aws:policy_document",
                    resource_id=policy_name,
                    attribute="inline_policies",
                )
            ]
            plan_response.extend(proposed_changes)

            if ctx.execute:
                log_str = f"{log_str} Removing inline policy..."

                apply_awaitable = boto_crud_call(
                    iam_client.delete_user_policy,
                    UserName=user_name,
                    PolicyName=policy_name,
                )
                delete_tasks.append(
                    plugin_apply_wrapper(apply_awaitable, proposed_changes)
                )

            log.debug(log_str, policy_name=policy_name, **log_params)

    for policy_name, policy_document in template_policy_map.items():
        existing_policy_doc = existing_policy_map.get(policy_name)
        policy_drift = None
        if existing_policy_doc:
            policy_drift = await aio_wrapper(
                DeepDiff,
                existing_policy_doc,
                policy_document,
                ignore_order=True,
                report_repetition=True,
                exclude_regex_paths=["metadata_commented_dict"],
            )

            # DeepDiff will return type changes as actual type functions and not strings,
            # and this will cause json serialization to fail later on when we process
            # the proposed changes. We force type changes to strings here.
            policy_drift = json.loads(policy_drift.to_json())

        if not existing_policy_doc or policy_drift:
            if policy_drift:
                log_params["policy_drift"] = policy_drift
                boto_action = "Updating"
                resource_existence = "Stale"
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.UPDATE,
                        resource_type="aws:policy_document",
                        resource_id=policy_name,
                        attribute="inline_policies",
                        change_summary=policy_drift,
                        current_value=existing_policy_doc,
                        new_value=policy_document,
                    )
                ]
            else:
                boto_action = "Creating"
                resource_existence = "New"
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.CREATE,
                        resource_type="aws:policy_document",
                        resource_id=policy_name,
                        attribute="inline_policies",
                        new_value=policy_document,
                    )
                ]
            plan_response.extend(proposed_changes)

            log_str = f"{resource_existence} inline policies discovered."
            if ctx.execute and policy_document:
                log_str = f"{log_str} {boto_action} inline policy..."

                apply_awaitable = boto_crud_call(
                    iam_client.put_user_policy,
                    UserName=user_name,
                    PolicyName=policy_name,
                    PolicyDocument=json.dumps(policy_document),
                )
                apply_tasks.append(
                    plugin_apply_wrapper(apply_awaitable, proposed_changes)
                )

            log.debug(log_str, policy_name=policy_name, **log_params)

    if apply_tasks or delete_tasks:
        results: list[list[ProposedChange]] = []
        if delete_tasks:
            results.extend(await asyncio.gather(*delete_tasks))
            if apply_tasks:
                # Wait for the policy deletion to propagate before applying new policies
                # Otherwise a max policy limit error may be thrown
                await asyncio.sleep(3)

        if apply_tasks:
            results.extend(await asyncio.gather(*apply_tasks))

        return list(chain.from_iterable(results))
    else:
        return plan_response


async def apply_user_groups(
    user_name,
    iam_client,
    template_groups: list[dict],
    existing_groups: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    tasks = []
    response = []
    template_groups = [group["GroupName"] for group in template_groups]
    existing_groups = [group["GroupName"] for group in existing_groups]

    # Create new groups
    for group in template_groups:
        if group not in existing_groups:
            log_str = "New groups discovered."
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.CREATE,
                    resource_type="aws:iam:group",
                    resource_id=group,
                    attribute="groups",
                )
            ]
            response.extend(proposed_changes)
            if ctx.execute:
                log_str = f"{log_str} Adding user to group..."
                apply_awaitable = boto_crud_call(
                    iam_client.add_user_to_group,
                    GroupName=group,
                    UserName=user_name,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))

            log.debug(log_str, group_name=group, **log_params)

    # Remove stale groups
    for group in existing_groups:
        if group not in template_groups:
            log_str = "Stale groups discovered."
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.DELETE,
                    resource_type="aws:iam:group",
                    resource_id=group,
                    attribute="groups",
                )
            ]
            response.extend(proposed_changes)
            if ctx.execute:
                log_str = f"{log_str} Removing user from group..."
                apply_awaitable = boto_crud_call(
                    iam_client.remove_user_from_group,
                    GroupName=group,
                    UserName=user_name,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))

            log.debug(log_str, group_name=group, **log_params)

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def delete_iam_user(user_name: str, iam_client, log_params: dict):
    tasks = []
    # Detach managed policies
    managed_policies = await get_user_managed_policies(user_name, iam_client)
    if managed_policies:
        managed_policies = [policy["PolicyArn"] for policy in managed_policies]
        log.debug(
            "Detaching managed policies.",
            managed_policies=managed_policies,
            **log_params,
        )
        for policy in managed_policies:
            tasks.append(
                boto_crud_call(
                    iam_client.detach_user_policy, UserName=user_name, PolicyArn=policy
                )
            )

    # Delete inline policies
    inline_policies = await get_user_inline_policies(user_name, iam_client)
    if inline_policies:
        inline_policies = list(inline_policies.keys())
        log.debug(
            "Deleting inline policies.", managed_policies=inline_policies, **log_params
        )
        for policy_name in inline_policies:
            tasks.append(
                boto_crud_call(
                    iam_client.delete_user_policy,
                    UserName=user_name,
                    PolicyName=policy_name,
                )
            )

    # Remove groups
    user_groups = await get_user_groups(user_name, iam_client)
    if user_groups:
        user_groups = sorted(list(user_groups.keys()))
        log.debug("Removing user groups.", groups=user_groups, **log_params)
        for group in user_groups:
            tasks.append(
                boto_crud_call(
                    iam_client.remove_user_from_group,
                    UserName=user_name,
                    GroupName=group,
                )
            )

    # Actually perform the deletion of Managed & Inline policies
    await asyncio.gather(*tasks)
    # Now that everything has been removed from the user, delete the user itself
    await boto_crud_call(iam_client.delete_user, UserName=user_name)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template,
    delete_orphaned_templates,
    get_existing_template_map,
)
from iambic.plugins.v0_1_0.azure_ad.group.models import (
    AZURE_AD_GROUP_TEMPLATE_TYPE,
    AzureActiveDirectoryGroupTemplate,
)
from iambic.plugins.v0_1_0.azure_ad.group.utils import list_groups

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.azure_ad.iambic_plugin import AzureADConfig


def get_resource_dir_args() -> list:
    return ["group"]


def get_response_dir(exe_message: ExecutionMessage) -> str:
    return exe_message.get_directory(*get_resource_dir_args(), "templates")


async def update_or_create_group_template(
    discovered_template: AzureActiveDirectoryGroupTemplate, existing_template_map: dict
) -> AzureActiveDirectoryGroupTemplate:
    """
    Update or create an AzureActiveDirectoryGroupTemplate object from the provided Group object.

    Args:
        discovered_template (Group): The Group template generated from the Azure AD cloud response.
        existing_template_map (dict): Existing IAMbic Azure AD group templates.
    """

    return create_or_update_template(
        discovered_template.file_path,
        existing_template_map,
        discovered_template.resource_id,
        AzureActiveDirectoryGroupTemplate,
        {"idp_name": discovered_template.idp_name},
        discovered_template.properties,
        [],
    )


async def collect_org_groups(exe_message: ExecutionMessage, config: AzureADConfig):
    assert exe_message.provider_id
    base_path = get_response_dir(exe_message)
    azure_organization = config.get_organization(exe_message.provider_id)
    log.info(
        "Beginning to retrieve Azure AD groups.", organization=exe_message.provider_id
    )

    groups = await list_groups(azure_organization)
    for group in groups:
        azure_group = AzureActiveDirectoryGroupTemplate(
            file_path="unset",
            idp_name=azure_organization.idp_name,
            properties=group,
        )
        azure_group.file_path = os.path.join(base_path, f"{group.name}.yaml")
        azure_group.write()

    log.info(
        "Finished retrieving Azure AD groups.",
        azure_org=exe_message.provider_id,
        group_count=len(groups),
    )


async def generate_group_templates(
    config: AzureADConfig,
    exe_message: ExecutionMessage,
    output_dir: str,
    detect_messages: list = None,
):
    """Create the templates for all collected groups in the domain"""
    base_path = os.path.expanduser(output_dir)
    existing_template_map = await get_existing_template_map(
        base_path,
        AZURE_AD_GROUP_TEMPLATE_TYPE,
        config.template_map,
    )
    all_resource_ids = set()

    log.info("Updating and creating Azure AD group templates.")

    groups = await exe_message.get_sub_exe_files(
        *get_resource_dir_args(), "templates", file_name_and_extension="**.yaml"
    )
    # Update or create templates
    for group in groups:
        group = AzureActiveDirectoryGroupTemplate(file_path="unset", **group)
        group.set_default_file_path(output_dir, group.properties.name)
        resource_template = await update_or_create_group_template(
            group, existing_template_map
        )
        if not resource_template:
            # Template not updated. Most likely because it's an `enforced` template.
            continue
        all_resource_ids.add(resource_template.resource_id)

    # Delete templates that no longer exist
    delete_orphaned_templates(list(existing_template_map.values()), all_resource_ids)

    log.info("Finished updating and creating Azure AD group templates.")



from __future__ import annotations

import asyncio
import os
from itertools import groupby
from typing import TYPE_CHECKING, Any, Optional, Union

import aiofiles

from iambic.core import noq_json as json
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template,
    delete_orphaned_templates,
)
from iambic.core.utils import NoqSemaphore, resource_file_upsert
from iambic.plugins.v0_1_0.aws.event_bridge.models import (
    SCPMessageDetails as SCPPolicyMessageDetails,
)
from iambic.plugins.v0_1_0.aws.models import AWSAccount, AWSOrganization
from iambic.plugins.v0_1_0.aws.organizations.scp.exceptions import (
    OrganizationAccountRequiredException,
)
from iambic.plugins.v0_1_0.aws.organizations.scp.models import (
    AWS_SCP_POLICY_TEMPLATE,
    AwsScpPolicyTemplate,
    PolicyProperties,
    ServiceControlPolicyCache,
    ServiceControlPolicyItem,
    ServiceControlPolicyResourceFiles,
)
from iambic.plugins.v0_1_0.aws.organizations.scp.utils import get_policy, list_policies
from iambic.plugins.v0_1_0.aws.utils import get_aws_account_map

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig

RESOURCE_DIR = ["organizations", "scp"]


def get_response_dir(exe_message: ExecutionMessage, aws_account: AWSAccount) -> str:
    if exe_message.provider_id:
        return exe_message.get_directory(*RESOURCE_DIR)
    else:
        return exe_message.get_directory(aws_account.account_id, *RESOURCE_DIR)


def get_template_dir(base_dir: str) -> str:
    return str(os.path.join(base_dir, "resources", "aws", *RESOURCE_DIR))


async def generate_scp_resource_files(
    exe_message: ExecutionMessage,
    aws_account: AWSAccount,
    filtered_policies: Optional[list[str]] = None,
) -> ServiceControlPolicyResourceFiles:
    resource_dir = get_response_dir(exe_message, aws_account)
    resource_file_upsert_semaphore = NoqSemaphore(resource_file_upsert, 10)
    messages = []

    response = ServiceControlPolicyResourceFiles(
        account_id=aws_account.account_id, policies=[]
    )
    organizations_client = await aws_account.get_boto3_client("organizations")
    if filtered_policies is None:
        scp_policies = await list_policies(organizations_client)
    elif len(filtered_policies) > 0:
        scp_policies = await asyncio.gather(
            *[
                get_policy(organizations_client, policy_id)
                for policy_id in set(filtered_policies)
            ]
        )
    else:
        return response

    log.debug(
        "Retrieved AWS SCP Policies.",
        account_id=aws_account.account_id,
        account_name=aws_account.account_name,
        scp_policies_count=len(scp_policies),
    )

    for policy in scp_policies:
        policy_path = os.path.join(resource_dir, f"{policy.Id}.json")
        response["policies"].append(
            ServiceControlPolicyCache(
                file_path=policy_path,
                policy_id=policy.Id,
                arn=policy.Arn,
                account_id=aws_account.account_id,
            )
        )
        messages.append(
            dict(
                file_path=policy_path,
                content_as_dict=policy.dict(),
                replace_file=True,
            )
        )

    await resource_file_upsert_semaphore.process(messages)
    log.debug(
        "Finished caching AWS SCP Policies.",
        account_id=aws_account.account_id,
        managed_policy_count=len(scp_policies),
    )

    return response


async def collect_aws_scp_policies(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    scp_template_map: dict,
    detect_messages: Optional[list[Union[SCPPolicyMessageDetails, Any]]] = None,
):
    aws_account_map = await get_organizations_account_map(exe_message, config)

    if detect_messages:
        detect_messages = [
            dm
            for dm in detect_messages
            if isinstance(dm, SCPPolicyMessageDetails)
            and exe_message.provider_id == dm.account_id
        ]
        if not detect_messages:
            return

    log.info(
        "Generating AWS SCP templates. Beginning to retrieve AWS SCP policies.",
        accounts=list(aws_account_map.keys()),
    )

    if detect_messages:
        # Remove deleted or mark templates for update
        delete_policies = set([dm.policy_id for dm in detect_messages if dm.delete])

        if delete_policies:
            existing_template_map = dict(
                [
                    (k, next(g))
                    for k, g in groupby(
                        scp_template_map.get(AWS_SCP_POLICY_TEMPLATE, {}).values(),
                        lambda p: p.properties.policy_id,
                    )
                ]
            )

            for policy in delete_policies:
                if existing_template := existing_template_map.get(policy):
                    existing_template.delete()

        generate_scp_resource_files_semaphore = NoqSemaphore(
            generate_scp_resource_files, 25
        )

        scp_policies: list[
            ServiceControlPolicyResourceFiles
        ] = await generate_scp_resource_files_semaphore.process(
            [
                {
                    "exe_message": exe_message,
                    "aws_account": aws_account_map[exe_message.provider_id],
                    "filtered_policies": set(
                        [
                            dm.policy_id
                            for dm in detect_messages
                            if not dm.delete and dm.policy_id not in delete_policies
                        ]
                    ),
                }
            ]
        )

    else:
        generate_scp_resource_files_semaphore = NoqSemaphore(
            generate_scp_resource_files, 25
        )

        scp_policies: list[
            ServiceControlPolicyResourceFiles
        ] = await generate_scp_resource_files_semaphore.process(
            [
                {
                    "exe_message": exe_message,
                    "aws_account": aws_account_map[exe_message.provider_id],  # type: ignore
                }
            ]
        )

    log.info(
        "Finished retrieving AWS SCP policy details",
        accounts=list(aws_account_map.keys()),
    )

    with open(
        exe_message.get_file_path(
            *RESOURCE_DIR,
            file_name_and_extension=f"output-{exe_message.provider_id}.json",
        ),
        "w",
    ) as f:
        f.write(json.dumps(scp_policies))


async def generate_aws_scp_policy_templates(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    base_output_dir: str,
    scp_template_map: dict,
    detect_messages: Optional[list[Union[SCPPolicyMessageDetails, Any]]] = None,
):
    """Generate AWS SCP policy templates

    Note:
        this function is executed for each organization account (provider_id)
        thus, each output file has one item in the list (scp_policies)
    """

    aws_account_map = await get_organizations_account_map(exe_message, config)

    existing_template_map = scp_template_map.get(AWS_SCP_POLICY_TEMPLATE, {})
    resource_dir = get_template_dir(base_output_dir)

    scp_policies: list[
        ServiceControlPolicyResourceFiles
    ] = await exe_message.get_sub_exe_files(
        *RESOURCE_DIR,
        file_name_and_extension=f"output-{exe_message.provider_id}.json",
        flatten_results=True,
    )  # type: ignore

    policies: list[Union[ServiceControlPolicyCache, dict]] = []
    account_id: str
    account_id, policies = scp_policies[0].values()  # type: ignore

    tasks = []
    organization: AWSOrganization = list(
        filter(lambda o: o.org_account_id == account_id, config.organizations)
    )[0]

    for policy in policies:
        tasks.append(
            upsert_templated_scp_policies(
                aws_account_map,
                account_id,
                policy,  # type: ignore
                resource_dir,
                existing_template_map,
                config,
                organization,
            )
        )

    log.debug("Writing templated scp policies")

    templates: list[AwsScpPolicyTemplate] = await asyncio.gather(*tasks)

    # NEVER call this if messages are passed in because all_resource_ids will only contain those resources
    if not detect_messages:
        # if some templates are iambic managed, they will be None
        all_resource_ids = set([t.identifier for t in templates if t is not None])
        delete_orphaned_templates(
            list(existing_template_map.values()), all_resource_ids
        )

    log.info("Finished templated scp policies generation")

    return


async def get_organizations_account_map(exe_message, config):
    aws_account_map = await get_aws_account_map(config)

    # SCP policies should be retrieved just for the organization account (provider_id)
    if not exe_message.provider_id or not aws_account_map.get(exe_message.provider_id):
        raise OrganizationAccountRequiredException()

    aws_account_map = {
        exe_message.provider_id: aws_account_map[exe_message.provider_id]
    }

    return aws_account_map


async def upsert_templated_scp_policies(
    aws_account_map: dict,
    account_id: str,
    policy: ServiceControlPolicyCache,  # type: ignore
    resource_dir: str,
    existing_template_map: dict,
    config,
    organization: AWSOrganization,
):
    async with aiofiles.open(policy.get("file_path"), mode="r") as f:
        content_dict = json.loads(await f.read())
        # policy = normalize_dict_keys(content_dict)  # type: ignore
        policy: ServiceControlPolicyItem = ServiceControlPolicyItem.parse_obj(
            content_dict
        )

    file_path = get_template_file_path(
        resource_dir,
        policy.Name,
        [aws_account_map[account_id].account_name],
        aws_account_map,
    )

    template_params, template_properties = AwsScpPolicyTemplate.factory_template_props(
        account_id,
        policy,
        config,
        organization,
    )

    return create_or_update_template(
        file_path,
        existing_template_map,
        policy.Name,
        AwsScpPolicyTemplate,
        template_params,
        PolicyProperties(**template_properties),
        list(aws_account_map.values()),
    )


def get_template_file_path(
    managed_policy_dir: str,
    policy_name: str,
    included_accounts: list[str],
    account_map: dict[str, AWSAccount],
):
    if len(included_accounts) > 1:
        separator = "multi_account"
    elif included_accounts == ["*"] or included_accounts is None:
        separator = "all_accounts"
    else:
        separator = included_accounts[0]

    file_name = (
        policy_name.replace(" ", "")
        .replace("{{var.", "")
        .replace("{{", "")
        .replace("}}_", "_")
        .replace("}}", "_")
        .replace(".", "_")
        .lower()
    )
    return str(os.path.join(managed_policy_dir, separator, f"{file_name}.yaml"))



from __future__ import annotations

from collections import defaultdict
from typing import Optional, Type

from iambic.core.models import BaseTemplate, ProviderChild
from iambic.core.utils import evaluate_on_provider


def group_detect_messages(group_by: str, messages: list) -> dict:
    """Group messages by a key in the message dict.

    Args:
        group_by (str): The key to group by.
        messages (list): The messages to group.

    Returns:
        dict: The grouped messages.
    """
    grouped_messages = defaultdict(list)
    for message in messages:
        grouped_messages[getattr(message, group_by)].append(message)

    return grouped_messages


def generate_template_output(
    excluded_provider_ids: list[str],
    provider_child_map: dict[str, ProviderChild],
    template: Optional[Type[BaseTemplate]],
) -> dict[str, dict]:
    provider_children_value_map = dict()
    if not template:
        return provider_children_value_map
    for provider_child_id, provider_child in provider_child_map.items():
        if provider_child_id in excluded_provider_ids:
            continue
        elif not evaluate_on_provider(
            template, provider_child, exclude_import_only=False
        ):
            continue

        if provider_child_value := template.apply_resource_dict(provider_child):
            provider_children_value_map[provider_child_id] = provider_child_value

    return provider_children_value_map



from __future__ import annotations

from pydantic import BaseModel

from iambic.core.iambic_plugin import ProviderPlugin
from iambic.core.models import ConfigMixin
from iambic.plugins.v0_1_0 import PLUGIN_VERSION
from iambic.plugins.v0_1_0.example.handlers import import_example_resources, load


def get_example_templates():
    from iambic.plugins.v0_1_0.example.local_database.models import (
        ExampleLocalDatabaseTemplate,
    )
    from iambic.plugins.v0_1_0.example.local_file.models import (
        ExampleLocalFileMultiAccountTemplate,
        ExampleLocalFileTemplate,
    )

    return [
        ExampleLocalFileTemplate,
        ExampleLocalFileMultiAccountTemplate,
        ExampleLocalDatabaseTemplate,
    ]


class ExampleConfig(ConfigMixin, BaseModel):
    @property
    def templates(self):
        return get_example_templates()


IAMBIC_PLUGIN = ProviderPlugin(
    config_name="example",
    version=PLUGIN_VERSION,
    provider_config=ExampleConfig,
    requires_secret=True,
    async_import_callable=import_example_resources,
    async_load_callable=load,
    templates=get_example_templates(),
)



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field


class PermissionSetMessageDetails(PydanticBaseModel):
    account_id: str
    instance_arn: str
    permission_set_arn: str


class ManagedPolicyMessageDetails(PydanticBaseModel):
    account_id: str
    policy_path: str
    policy_name: str
    delete: bool


class RoleMessageDetails(PydanticBaseModel):
    account_id: str
    role_name: str
    delete: bool


class GroupMessageDetails(PydanticBaseModel):
    account_id: str
    group_name: str
    delete: bool


class UserMessageDetails(PydanticBaseModel):
    account_id: str
    user_name: str
    delete: bool


class SCPMessageDetails(PydanticBaseModel):
    account_id: str
    policy_id: str
    delete: bool
    event: str = Field(
        ...,
        description="One of: CreatePolicy, DeletePolicy, UpdatePolicy, AttachPolicy, DetachPolicy, TagResource, UntagResource",
    )

    @staticmethod
    def tag_event(event, source) -> bool:
        """Returns True if the event is a tag/untag event related to SCPs"""
        return (
            event in ["TagResource", "UntagResource"]
            and source == "organizations.amazonaws.com"
        )

    @staticmethod
    def get_policy_id(request_params, response_elements) -> Optional[str]:
        """Returns the policy ID from the request parameters or response elements (last one if it is a CreatePolicy event)"""
        return (request_params and request_params.get("policyId", None)) or (
            response_elements
            and response_elements.get("policy", {})
            .get("policySummary", {})
            .get("id", None)
        )



from __future__ import annotations

import asyncio
import json
from enum import Enum
from itertools import chain
from typing import TYPE_CHECKING, Callable, List, Optional, TypedDict, Union

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, validator

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
)
from iambic.core.utils import normalize_dict_keys, plugin_apply_wrapper
from iambic.plugins.v0_1_0.aws.iam.models import Path
from iambic.plugins.v0_1_0.aws.iam.policy.models import PolicyStatement
from iambic.plugins.v0_1_0.aws.models import (
    AccessModel,
    AWSAccount,
    AWSTemplate,
    Description,
    StatementEffect,
    Tag,
)
from iambic.plugins.v0_1_0.aws.organizations.scp.utils import (
    apply_update_policy,
    apply_update_policy_tags,
    apply_update_policy_targets,
    create_policy,
    delete_policy,
    get_policy,
    service_control_policy_is_enabled,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig
    from iambic.plugins.v0_1_0.aws.models import AWSOrganization


AWS_SCP_POLICY_TEMPLATE = "NOQ::AWS::Organizations::SCP"


class ServiceControlPolicyTargetItemType(str, Enum):
    ACCOUNT = "ACCOUNT"
    ORGANIZATIONAL_UNIT = "ORGANIZATIONAL_UNIT"
    ROOT = "ROOT"


class OrganizationsPolicyType(str, Enum):
    """
    AWS Organizations supports the following policy types. You specify the policy type when you create a policy.

    Possible values:
        - TAG_POLICY
        - BACKUP_POLICY
        - AISERVICES_OPT_OUT_POLICY

    Ref:
        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/organizations/client/list_policies.html
    """

    SERVICE_CONTROL_POLICY = "SERVICE_CONTROL_POLICY"
    TAG_POLICY = "TAG_POLICY"
    BACKUP_POLICY = "BACKUP_POLICY"
    AISERVICES_OPT_OUT_POLICY = "AISERVICES_OPT_OUT_POLICY"


class OutputModels(PydanticBaseModel):
    """Response models from AWS Organizations client"""

    class Config:
        use_enum_values = True


class PolicyTargetProperties(BaseModel):
    """
    Note:
    - Root - A string that begins with “r-” followed by from 4 to 32 lowercase letters or digits.
    - Account - A string that consists of exactly 12 digits.
    - Organizational unit (OU) - A string that begins with “ou-” followed by from 4 to 32 lowercase letters or digits (the ID of the root
    that the OU is in). This string is followed by a second “-” dash and from 8 to 32 additional lowercase letters or digits.
    """

    organizational_units: list[str] = Field(default=[], description="List of OUs ids")
    accounts: list[str] = Field(
        default=[], description="List of accounts (names or ids)"
    )
    roots: list[str] = Field(default=[], description="List of root ids")

    @property
    def resource_type(self):
        return "aws:iam:scp_policy:target"

    @property
    def resource_id(self):
        return "|".join(self.organizational_units + self.accounts + self.roots)

    @staticmethod
    def parse_targets(targets: list, config: AWSConfig):
        data = dict(organizational_units=[], accounts=[], roots=[])

        for target in targets:
            key = "accounts"
            target_id = target.get("target_id")
            if target_id.startswith("o-") or target_id.startswith("ou-"):
                key = "organizational_units"
            elif target_id.startswith("r-"):
                key = "roots"
            else:
                target_id = list(
                    map(
                        lambda a: a.account_name,
                        filter(lambda a: a.account_id == target_id, config.accounts),
                    )
                )[0]

            data[key].append(target_id)

        return data

    @staticmethod
    def unparse_targets(targets: list[str], config: AWSConfig):
        data = []

        for target in targets:
            if (
                target.startswith("o-")
                or target.startswith("ou-")
                or target.startswith("r-")
            ):
                target_id = target
            elif target.isdigit() and len(target) == 12:
                # warn: it could be possible that a person has an account name with 12 digits
                target_id = target
            else:
                target_id = list(
                    map(
                        lambda a: a.account_id,
                        filter(lambda a: a.account_name == target, config.accounts),
                    )
                )[0]

            data.append(target_id)

        return data


class PolicyDocument(AccessModel, ExpiryModel):
    version: str = "2012-10-17"
    statement: Union[List[PolicyStatement], PolicyStatement] = Field(
        ...,
        description="List of policy statements",
    )

    @property
    def resource_type(self):
        return "aws:policy_document"

    @property
    def resource_id(self):
        return self.statement.sid

    @staticmethod
    def parse_raw_policy(resource_raw) -> "PolicyDocument":
        resource = json.loads(resource_raw)
        resource = normalize_dict_keys(resource)
        return PolicyDocument(**resource)  # type: ignore

    @validator("statement")
    def validate_statement(cls, statements):
        keys = ["principal", "not_principal", "not_resource"]
        for key in keys:
            for statement in statements:
                if getattr(statement, key, None):
                    raise ValueError(f"{key} is not supported")
        return statements


class PolicyProperties(BaseModel):
    policy_id: Optional[str] = Field(
        None,
        description="The ID of the policy, it is optional when creating a new policy",
        required=False,
    )
    policy_name: str
    path: Optional[Union[str, List["Path"]]] = "/"
    description: Optional[Union[str, list[Description]]] = Field(
        "",
        description="Description of the role",
    )
    type: OrganizationsPolicyType = Field(
        default=OrganizationsPolicyType.SERVICE_CONTROL_POLICY
    )
    aws_managed: bool = Field(False)
    policy_document: Union[PolicyDocument, List[PolicyDocument]] = Field(
        ...,
        description="Policy document, Unsupported elements: Principal, NotPrincipal, NotResource",
    )
    targets: PolicyTargetProperties = Field(default=None)

    tags: Optional[List[Tag]] = Field(
        [],
        description="List of tags attached to the role",
    )

    @property
    def resource_type(self):
        return "aws:iam:scp_policy:properties"

    @property
    def resource_id(self):
        return self.policy_name

    @validator("policy_document")
    def sort_policy_document(cls, v: list[PolicyDocument]):
        if not isinstance(v, list):
            return v
        sorted_v = sorted(v, key=lambda d: d.access_model_sort_weight())
        return sorted_v

    @classmethod
    def sort_func(cls, attribute_name: str) -> Callable:
        def _sort_func(obj):
            return f"{getattr(obj, attribute_name)}!{obj.access_model_sort_weight()}"

        return _sort_func

    @validator("tags")
    def sort_tags(cls, v: list[Tag]) -> list[Tag]:
        sorted_v = sorted(v, key=cls.sort_func("key"))
        return sorted_v


class AwsScpPolicyTemplate(AWSTemplate, AccessModel):
    template_type = AWS_SCP_POLICY_TEMPLATE
    organization_account_needed: bool = Field(
        True,
        description="This template needs an organization account to be applied",
    )
    ARN_TEMPLATE = "arn:aws:organizations::{account_id}:policy/{organization_unit}/service_control_policy/{policy_id}"

    properties: PolicyProperties = Field(
        description="The properties of the scp policy",
    )
    account_id: str
    org_id: str

    def get_arn(self) -> str:
        return self.ARN_TEMPLATE.format(
            account_id=self.account_id if not self.properties.aws_managed else "aws",
            organization_unit=self.org_id,
            policy_id=self.properties.policy_id,
        )

    def _apply_resource_dict(self, aws_account: AWSAccount = None) -> dict:
        resource_dict = super()._apply_resource_dict(aws_account)
        resource_dict["Arn"] = self.get_arn()
        return resource_dict

    async def _apply_to_account(self, aws_account: AWSAccount) -> AccountChangeDetails:
        if self.account_id != aws_account.account_id:
            return AccountChangeDetails(
                account=str(aws_account),
                resource_id=self.properties.resource_id,
                resource_type=self.resource_type,
                new_value={},
                current_value={},
                proposed_changes=[],
                exceptions_seen=[],
            )  # type: ignore

        client = await aws_account.get_boto3_client("organizations")

        if not (await service_control_policy_is_enabled(client)):
            log.info("Service control policy is not enabled in the organization")
            return AccountChangeDetails(
                account=str(aws_account),
                resource_id=self.properties.resource_id,
                resource_type=self.resource_type,
                new_value={},
                current_value={},
                proposed_changes=[],
                exceptions_seen=[],
            )  # type: ignore

        account_policy = self.apply_resource_dict(aws_account)
        policy_name = account_policy.get("PolicyName", "")

        account_change_details = AccountChangeDetails(
            account=str(aws_account),
            resource_id=self.properties.resource_id,
            resource_type=self.resource_type,
            new_value=dict(**account_policy),
            proposed_changes=[],
            exceptions_seen=[],
        )  # type: ignore

        log_params = dict(
            resource_type=self.resource_type,
            resource_id=self.properties.resource_id,
            account=str(aws_account),
        )

        current_policy = None
        if account_policy.get("PolicyId"):
            current_policy = await get_policy(client, account_policy.get("PolicyId"))
            current_policy = current_policy.dict()

        if current_policy:
            # UPDATE POLICY
            account_change_details.current_value = {**current_policy}

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                account_change_details.new_value = {}
                return account_change_details

        if self.is_delete_action(aws_account):
            # DELETE POLICY
            if current_policy:
                account_change_details.new_value = None
                proposed_changes = [
                    ProposedChange(
                        change_type=ProposedChangeType.DELETE,
                        resource_id=self.properties.resource_id,
                        resource_type=self.resource_type,
                    )  # type: ignore
                ]
                log_str = "Active resource found with deleted=false."
                if ctx.execute:
                    log_str = f"{log_str} Deleting resource..."
                log.debug(log_str, **log_params)

                if ctx.execute:
                    apply_awaitable = delete_policy(
                        client, account_policy.get("PolicyId"), log_params
                    )
                    proposed_changes = await plugin_apply_wrapper(
                        apply_awaitable, proposed_changes
                    )

                account_change_details.extend_changes(proposed_changes)

            return account_change_details

        if current_policy:
            args = [
                client,
                account_policy,
                current_policy,
                log_params,
                aws_account,
            ]

            tasks = [
                method(*args)
                for method in (
                    apply_update_policy,
                    apply_update_policy_targets,
                    apply_update_policy_tags,
                )
            ]

            changes_made: list[list[ProposedChange]] = await asyncio.gather(*tasks)
            if any(changes_made):
                account_change_details.extend_changes(
                    list(chain.from_iterable(changes_made))
                )
                if current_policy.get("Name") != account_policy.get("PolicyName"):
                    self.identifier = account_policy["PolicyName"]
                    self.write()
        else:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.CREATE,
                    resource_id=policy_name,
                    resource_type=self.resource_type,
                    change_summary=account_policy,
                )  # type: ignore
            ]
            log_str = "New resource found in code."
            if not ctx.execute:
                # Exit now because apply functions won't work if resource doesn't exist
                log.debug(log_str, **log_params)
                account_change_details.extend_changes(proposed_changes)
                return account_change_details

            log.debug(f"{log_str} Creating resource...", **log_params)

            apply_awaitable = create_policy(client, account_policy)
            account_change_details.extend_changes(
                await plugin_apply_wrapper(apply_awaitable, proposed_changes)
            )

            self.properties.policy_id = account_policy.get("PolicyId")
            current_policy = await get_policy(client, account_policy.get("PolicyId"))
            current_policy = current_policy.dict()

            args = [client, account_policy, current_policy, log_params, aws_account]

            tasks = [
                apply_update_policy_tags(*args),
                apply_update_policy_targets(*args),
            ]

            changes_made: list[list[ProposedChange]] = await asyncio.gather(*tasks)
            if any(changes_made):
                account_change_details.extend_changes(
                    list(chain.from_iterable(changes_made))
                )

            # name and identifier must match
            self.identifier = current_policy.get("Name", self.identifier)
            self.write()

        self.__log_after_apply(account_change_details, log_params)

        return account_change_details

    def is_delete_action(self, aws_account):
        deleted = self.get_attribute_val_for_account(aws_account, "deleted", False)
        if isinstance(deleted, list):
            deleted = deleted[0].deleted

        return deleted

    def __log_after_apply(self, account_change_details, log_params):
        if ctx.execute and not account_change_details.exceptions_seen:
            log.debug(
                "Successfully finished execution on account for resource",
                changes_made=bool(account_change_details.proposed_changes),
                **log_params,
            )
        elif account_change_details.exceptions_seen:
            log.error(
                "Unable to finish execution on account for resource",
                exceptions_seen=[
                    cd.exceptions_seen for cd in account_change_details.exceptions_seen
                ],
                **log_params,
            )
        else:
            log.debug(
                "Successfully finished scanning for drift on account for resource",
                requires_changes=bool(account_change_details.proposed_changes),
                **log_params,
            )

    @staticmethod
    def factory_template_props(
        account_id: str,
        policy: ServiceControlPolicyItem,
        config: AWSConfig,
        organization: AWSOrganization,
    ):
        template_params = dict(
            identifier=policy.Name,
            account_id=account_id,
            org_id=organization.org_id,
        )

        template_properties = dict(
            policy_id=policy.Id,
            policy_name=policy.Name,
            description=policy.Description,
            type=policy.Type,
            aws_managed=policy.AwsManaged,
            policy_document=dict(
                version=policy.PolicyDocument.Version,
                statement=policy.PolicyDocument.Statement,
            ),
        )

        if policy.Targets:
            template_properties.update(
                targets=PolicyTargetProperties.parse_obj(
                    PolicyTargetProperties.parse_targets(
                        [normalize_dict_keys(t.dict()) for t in policy.Targets],
                        config,
                    )
                )  # type: ignore
            )

        if policy.Tags:
            template_properties.update(
                tags=[normalize_dict_keys(t.dict()) for t in policy.Tags]  # type: ignore
            )
        return template_params, template_properties


class PolicyStatementItem(OutputModels):
    Sid: Optional[str]
    Effect: StatementEffect
    Action: Optional[Union[list[str], str]]
    NotAction: Optional[Union[list[str], str]]
    Resource: Union[list[str], str]
    Condition: Optional[dict]


class PolicyDocumentItem(OutputModels):
    Version: str
    Statement: Union[list[PolicyStatementItem], PolicyStatementItem]


class ServiceControlPolicyTargetItem(OutputModels):
    TargetId: str = Field(
        description="""
Root - A string that begins with “r-” followed by from 4 to 32 lowercase
letters or digits.
Account - A string that consists of exactly 12 digits.
Organizational unit (OU) - A string that begins with “ou-” followed by
from 4 to 32 lowercase letters or digits (the ID of the root
that the OU is in). This string is followed by a second “-” dash
and from 8 to 32 additional lowercase letters or digits.
    """
    )
    Arn: str
    Name: str
    Type: ServiceControlPolicyTargetItemType = Field(
        default=ServiceControlPolicyTargetItemType.ACCOUNT
    )


class TagItem(OutputModels):
    Key: str
    Value: str


class ServiceControlPolicyItem(OutputModels):
    Id: str
    Arn: str
    Name: str
    Description: str
    Type: OrganizationsPolicyType = Field(
        default=OrganizationsPolicyType.SERVICE_CONTROL_POLICY
    )
    AwsManaged: bool
    Targets: list[ServiceControlPolicyTargetItem] = Field(default_factory=list)
    PolicyDocument: PolicyDocumentItem
    Tags: List[TagItem]


class ServiceControlPolicyCache(TypedDict):
    file_path: str
    policy_id: str
    arn: str
    account_id: str


class ServiceControlPolicyResourceFiles(TypedDict):
    account_id: str
    policies: list[ServiceControlPolicyCache]



from __future__ import annotations

IAMBIC_APPLY_ERROR_METADATA = "<!--iambic/apply/error-->"


def is_iambic_apply_error(body: str):
    footer = body.split("\n")[-1]
    return IAMBIC_APPLY_ERROR_METADATA in footer



from __future__ import annotations

import asyncio
import json
import random
from itertools import chain

from deepdiff import DeepDiff
from git import TYPE_CHECKING
from tenacity import retry, stop_after_attempt, wait_exponential

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import NoqSemaphore, aio_wrapper, plugin_apply_wrapper
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call, legacy_paginated_search

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.models import AWSAccount
    from iambic.plugins.v0_1_0.aws.organizations.scp.models import (
        ServiceControlPolicyItem,
        ServiceControlPolicyTargetItem,
    )


async def list_policies(
    client, filter="SERVICE_CONTROL_POLICY"
) -> list[ServiceControlPolicyItem]:
    """Retrieves the list of all policies in an organization of a specified type."""

    from iambic.plugins.v0_1_0.aws.organizations.scp.models import (
        ServiceControlPolicyItem,
    )

    scp_policies = await legacy_paginated_search(
        client.list_policies,
        response_key="Policies",
        **dict(Filter=filter),
    )

    scp_policies = [p for p in scp_policies if p["AwsManaged"] is False]

    list_targets_for_policy_semaphore = NoqSemaphore(list_targets_for_policy, 10)
    describe_policy_semaphore = NoqSemaphore(get_policy_statements, 10)
    list_tags_by_policy_semaphore = NoqSemaphore(list_tags_by_policy, 10)

    targets = await list_targets_for_policy_semaphore.process(
        [{"client": client, "policyId": policy["Id"]} for policy in scp_policies]
    )

    statements = await describe_policy_semaphore.process(
        [{"client": client, "policyId": policy["Id"]} for policy in scp_policies]
    )

    tags = await list_tags_by_policy_semaphore.process(
        [{"client": client, "policyId": policy["Id"]} for policy in scp_policies]
    )

    return [
        ServiceControlPolicyItem.parse_obj(
            {
                **p,
                "Targets": t,
                "PolicyDocument": s,
                "Tags": tg,
            },
        )
        for p, t, s, tg in zip(scp_policies, targets, statements, tags)
    ]


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
async def list_targets_for_policy(
    client, policyId: str
) -> list[ServiceControlPolicyTargetItem]:
    """
    Lists all the roots, organizational units (OUs),
    and accounts that the specified policy is attached to.
    """
    from iambic.plugins.v0_1_0.aws.organizations.scp.models import (
        ServiceControlPolicyTargetItem,
    )

    targets = await legacy_paginated_search(
        client.list_targets_for_policy,
        response_key="Targets",
        **dict(PolicyId=policyId),
    )

    return [ServiceControlPolicyTargetItem.parse_obj(t) for t in targets]


async def get_policy_statements(client, policyId: str):
    policy = await describe_policy(client, policyId)
    return policy.get("Content", {})


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
async def describe_policy(client, policyId: str):
    from iambic.plugins.v0_1_0.aws.organizations.scp.models import PolicyDocumentItem

    res = await boto_crud_call(client.describe_policy, PolicyId=policyId)
    return dict(
        PolicySummary=res.get("Policy", {}).get("PolicySummary", {}),
        Content=PolicyDocumentItem.parse_obj(
            json.loads(res.get("Policy", {}).get("Content", {}))
        ),
    )


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
async def list_tags_by_policy(client, policyId: str) -> list[dict]:
    """
    Lists tags that are attached to the specified resource.
    """
    targets = await legacy_paginated_search(
        client.list_tags_for_resource,
        response_key="Tags",
        **dict(ResourceId=policyId),
    )

    return targets


async def get_policy(client, policyId: str) -> ServiceControlPolicyItem:
    from iambic.plugins.v0_1_0.aws.organizations.scp.models import (
        ServiceControlPolicyItem,
    )

    policy = await describe_policy(client, policyId)

    list_tags_by_policy_semaphore = NoqSemaphore(list_tags_by_policy, 10)
    list_targets_for_policy_semaphore = NoqSemaphore(list_targets_for_policy, 10)

    tags = await list_tags_by_policy_semaphore.process(
        [{"client": client, "policyId": policyId}]
    )

    targets = await list_targets_for_policy_semaphore.process(
        [{"client": client, "policyId": policyId}]
    )

    return ServiceControlPolicyItem.parse_obj(
        {
            **policy.get("PolicySummary", {}),  # type: ignore
            "Targets": targets[0],
            "PolicyDocument": policy.get("Content", {}),
            "Tags": tags[0],
        },
    )


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
async def detach_policy(client, policyId, targetId):
    await boto_crud_call(
        client.detach_policy,
        PolicyId=policyId,
        TargetId=targetId,
    )
    log.debug(f"Detached policy {policyId} from {targetId}")


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=5, max=20),
)
async def delete_policy(client, policyId: str, *args, **kwargs):
    """
    Deletes the specified policy from your organization.
    Before you perform this operation, you must first detach
    the policy from all organizational units (OUs), roots, and accounts.
    """
    list_targets_for_policy_semaphore = NoqSemaphore(list_targets_for_policy, 10)

    targets = chain.from_iterable(
        await list_targets_for_policy_semaphore.process(
            [{"client": client, "policyId": policyId}]
        )
    )

    targets_tasks = [
        detach_policy(client, policyId, target.TargetId) for target in targets
    ]

    await asyncio.gather(*targets_tasks)

    await boto_crud_call(client.delete_policy, PolicyId=policyId)

    log.debug(f"Deleted policy {policyId}")


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
async def create_policy(client, policy):
    if isinstance(policy["PolicyDocument"], dict):
        policy["PolicyDocument"] = json.dumps(policy["PolicyDocument"])

    res = await boto_crud_call(
        client.create_policy,
        Content=policy["PolicyDocument"],
        Description=policy.get("Description", ""),
        Name=policy.get("PolicyName", f"NewPolicy-{random.randint(0, 100):03d}"),
        Type="SERVICE_CONTROL_POLICY",
    )

    policy.update(
        PolicyId=res.get("Policy", {}).get("PolicySummary", {}).get("Id"),
        PolicyName=res.get("Policy", {}).get("PolicySummary", {}).get("Name"),
    )

    return res.get("Policy", {}).get("PolicySummary", {})


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=5, max=20),
)
async def describe_organization(
    client,
):
    """Retrieves information about the organization that the user's account belongs to.
    This operation can be called from any account in the organization.
    """
    res = await boto_crud_call(client.describe_organization)
    return res.get("Organization", {})


async def service_control_policy_is_enabled(client):
    """Check if SCPs are enabled for the organization."""
    org = await describe_organization(client)

    return (
        len(
            [
                apt
                for apt in org.get("AvailablePolicyTypes", [])
                if apt.get("Type") == "SERVICE_CONTROL_POLICY"
                and apt.get("Status") == "ENABLED"
            ]
        )
        > 0
    )


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
async def apply_update_policy(
    client,
    policy,
    current_policy,
    log_params,
    *args,
) -> list[ProposedChange]:
    """Apply update policy.

    Args:
        client: organization client
        policy: policy read from yaml
        current_policy (dict): Response from boto3, based on ServiceControlPolicyItem.
        log_params (dict):

    Returns:
        list[ProposedChange]: list of proposed changes
    """

    from iambic.plugins.v0_1_0.aws.organizations.scp.models import PolicyDocumentItem

    current_value = dict(
        Name=current_policy.get("Name"),
        Description=current_policy.get("Description"),
        Content=current_policy["PolicyDocument"],
    )

    new_value = dict(
        Name=policy.get("PolicyName"),
        Description=policy.get("Description"),
        Content=PolicyDocumentItem.parse_obj(policy["PolicyDocument"]).dict(),
    )

    diff = await aio_wrapper(
        DeepDiff,
        current_value,
        new_value,
        report_repetition=True,
        ignore_order=True,
        exclude_regex_paths=["metadata_commented_dict"],
    )

    if diff == {}:
        return []

    tasks = []
    policy = policy.copy()

    if isinstance(policy["PolicyDocument"], dict):
        policy["PolicyDocument"] = json.dumps(policy["PolicyDocument"])

    proposed_changes = [
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_type=log_params.get("resource_type"),
            resource_id=policy.get("Name"),
            change_summary=diff.to_dict(),
            current_value=current_value,
            new_value=new_value,
        )  # type: ignore
    ]

    if ctx.execute:
        apply_awaitable = boto_crud_call(
            client.update_policy,
            PolicyId=policy.get("PolicyId"),
            Name=policy.get("PolicyName", f"NewPolicy-{random.randint(0, 100):03d}"),
            Description=policy.get("Description", ""),
            Content=policy["PolicyDocument"],
        )

        tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return proposed_changes


async def apply_update_policy_targets(
    client,
    policy,
    current_policy,
    log_params,
    aws_account: AWSAccount,
    *args,
) -> list[ProposedChange]:
    """Apply update policy targets.

    Args:
        client: organization client
        policy: policy read from yaml
        current_policy (dict): Response from boto3, based on ServiceControlPolicyItem.
        log_params (dict):

    Returns:
        list[ProposedChange]: list of proposed changes
    """
    tasks = []
    response = []

    t, r = __remove_targets(
        client,
        policy,
        current_policy,
        log_params,
        aws_account,
    )

    tasks += t
    response += r

    t, r = __apply_targets(
        client,
        policy,
        current_policy,
        log_params,
        aws_account,
    )

    tasks += t
    response += r

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def apply_update_policy_tags(
    client,
    policy,
    current_policy,
    log_params,
    *args,
) -> list[ProposedChange]:
    """Apply update policy tags.

    Args:
        client: organization client
        policy: policy read from yaml
        current_policy (dict): Response from boto3, based on ServiceControlPolicyItem.
        log_params (dict):

    Returns:
        list[ProposedChange]: list of proposed changes
    """

    tasks = []
    response = []

    t, r = __remove_tags(
        client,
        policy,
        current_policy,
        log_params,
    )

    tasks += t
    response += r

    t, r = __apply_tags(
        client,
        policy,
        current_policy,
        log_params,
    )

    tasks += t
    response += r

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
def __apply_tags(
    client,
    policy,
    current_policy,
    log_params,
):
    """Apply tags to policy"""
    response = []
    tasks = []
    existing_tag_map = {
        tag["Key"]: tag.get("Value") for tag in current_policy.get("Tags", [])
    }

    tags_to_apply = [
        tag
        for tag in policy.get("Tags", [])
        if tag.get("Value") != existing_tag_map.get(tag["Key"])
    ]

    if tags_to_apply:
        log_str = "New tags discovered in AWS."
        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_type=log_params.get("resource_type"),
                resource_id=policy.get("Name"),
                attribute="tags",
                current_value=[],
                new_value=tags_to_apply,
            )  # type: ignore
        ]
        response.extend(proposed_changes)
        if ctx.execute:
            log_str = f"{log_str} Adding tags..."
            apply_awaitable = boto_crud_call(
                client.tag_resource,
                ResourceId=policy.get("PolicyId"),
                Tags=tags_to_apply,
            )
            tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, tags=tags_to_apply, **log_params)

    return tasks, response


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
def __remove_tags(client, policy, current_policy, log_params):
    """Remove tags from the policy that are not in the template."""
    response = []
    tasks = []
    template_tag_map = {tag["Key"]: tag.get("Value") for tag in policy.get("Tags", [])}

    if tags_to_remove := [
        tag["Key"]
        for tag in current_policy.get("Tags", [])
        if not template_tag_map.get(tag["Key"])
    ]:
        log_str = "Stale tags discovered."
        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                attribute="tags",
                resource_type=log_params.get("resource_type"),
                resource_id=policy.get("Name"),
                current_value=tags_to_remove,
                new_value=[],
                change_summary={"TagKeys": tags_to_remove},
            )  # type: ignore
        ]
        response.extend(proposed_changes)

        if ctx.execute:
            log_str = f"{log_str} Removing tags..."
            apply_awaitable = boto_crud_call(
                client.untag_resource,
                ResourceId=policy.get("PolicyId"),
                TagKeys=tags_to_remove,
            )
            tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, tags=tags_to_remove, **log_params)

    return tasks, response


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
def __apply_targets(
    client,
    policy,
    current_policy,
    log_params,
    aws_account: AWSAccount,
):
    """Apply targets to policy."""
    from iambic.plugins.v0_1_0.aws.organizations.scp.models import (
        PolicyTargetProperties,
    )

    response = []
    tasks = []

    targets = list(
        chain.from_iterable(
            policy.get(
                "Targets", dict(OrganizationalUnits=[], Accounts=[], Roots=[])
            ).values()
        )
    )
    targets = PolicyTargetProperties.unparse_targets(targets, aws_account.aws_config)

    current_targets = list(
        map(lambda t: t.get("TargetId"), current_policy.get("Targets"))
    )

    if targets_to_apply := [tag for tag in targets if tag not in current_targets]:
        log_str = "New targets discovered."
        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_type=log_params.get("resource_type"),
                resource_id=policy.get("Name"),
                attribute="targets",
                current_value=current_targets,
                new_value=current_targets + targets_to_apply,
            )  # type: ignore
        ]
        response.extend(proposed_changes)
        if ctx.execute:
            log_str = f"{log_str} Adding targets..."

            async def attach_policies():
                """Due to ConcurrentModificationException, we need to execute it one at a time."""
                for target in targets_to_apply:
                    await boto_crud_call(
                        client.attach_policy,
                        PolicyId=policy.get("PolicyId"),
                        TargetId=target,
                    )

            tasks.append(plugin_apply_wrapper(attach_policies(), proposed_changes))

        log.debug(log_str, tags=targets_to_apply, **log_params)

    return tasks, response


@retry(
    reraise=True,
    stop=stop_after_attempt(6),
    wait=wait_exponential(multiplier=1, min=4, max=15),
)
def __remove_targets(
    client,
    policy,
    current_policy,
    log_params,
    aws_account: AWSAccount,
):
    """Remove targets from policy that are not in the template."""
    from iambic.plugins.v0_1_0.aws.organizations.scp.models import (
        PolicyTargetProperties,
    )

    response = []
    tasks = []

    targets = list(
        chain.from_iterable(
            policy.get(
                "Targets", dict(OrganizationalUnits=[], Accounts=[], Roots=[])
            ).values()
        )
    )
    targets = PolicyTargetProperties.unparse_targets(targets, aws_account.aws_config)

    current_targets = list(
        map(lambda t: t.get("TargetId"), current_policy.get("Targets"))
    )

    if targets_to_remove := [
        target for target in current_targets if target not in targets
    ]:
        log_str = "Stale targets discovered."
        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                attribute="targets",
                resource_type=log_params.get("resource_type"),
                resource_id=policy.get("Name"),
                current_value=targets_to_remove,
                new_value=[],
                change_summary={"Targets": targets_to_remove},
            )  # type: ignore
        ]
        response.extend(proposed_changes)

        if ctx.execute:
            log_str = f"{log_str} Removing targets..."

            async def detach_policies():
                """Due to ConcurrentModificationException, we need to execute it one at a time."""
                for target in targets_to_remove:
                    await boto_crud_call(
                        client.detach_policy,
                        PolicyId=policy.get("PolicyId"),
                        TargetId=target,
                    )

            tasks.append(plugin_apply_wrapper(detach_policies(), proposed_changes))
        log.debug(log_str, tags=targets_to_remove, **log_params)

    return tasks, response



from __future__ import annotations

import asyncio
import itertools
import os
from collections import defaultdict
from typing import TYPE_CHECKING, Optional

import aiofiles

from iambic.core import noq_json as json
from iambic.core.detect import generate_template_output, group_detect_messages
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template,
    delete_orphaned_templates,
)
from iambic.core.utils import NoqSemaphore, normalize_dict_keys, resource_file_upsert
from iambic.plugins.v0_1_0.aws.event_bridge.models import UserMessageDetails
from iambic.plugins.v0_1_0.aws.iam.user.models import (
    AWS_IAM_USER_TEMPLATE_TYPE,
    AwsIamUserTemplate,
    UserProperties,
)
from iambic.plugins.v0_1_0.aws.iam.user.utils import (
    get_user,
    get_user_groups,
    get_user_inline_policies,
    get_user_managed_policies,
    list_user_tags,
    list_users,
)
from iambic.plugins.v0_1_0.aws.models import AWSAccount
from iambic.plugins.v0_1_0.aws.template_generation import (
    base_group_str_attribute,
    group_dict_attribute,
    group_int_or_str_attribute,
)
from iambic.plugins.v0_1_0.aws.utils import (
    calculate_import_preference,
    get_aws_account_map,
    process_import_rules,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig

RESOURCE_DIR = ["iam", "user"]


def get_response_dir(exe_message: ExecutionMessage, aws_account: AWSAccount) -> str:
    if exe_message.provider_id:
        return exe_message.get_directory(*RESOURCE_DIR)
    else:
        return exe_message.get_directory(aws_account.account_id, *RESOURCE_DIR)


def get_template_dir(base_dir: str) -> str:
    return str(os.path.join(base_dir, "resources", "aws", *RESOURCE_DIR))


def get_templated_user_file_path(
    user_dir: str,
    user_name: str,
    included_accounts: list[str],
    user_path: str = None,
) -> str:
    if included_accounts is not None and len(included_accounts) > 1:
        separator = "multi_account"
    elif included_accounts == ["*"] or included_accounts is None:
        separator = "all_accounts"
    else:
        separator = included_accounts[0]

    file_name = (
        user_name.replace(" ", "")
        .replace("{{var.", "")
        .replace("{{", "")
        .replace("}}_", "_")
        .replace("}}", "_")
        .replace(".", "_")
        .lower()
    )

    # using path components from path attribute
    if user_path:
        user_path_components = user_path.split("/")
        # get rid of empty component
        user_path_components = [
            component for component in user_path_components if component
        ]

    # stitch desired location together
    os_paths = [user_dir, separator]
    os_paths.extend(user_path_components)
    os_paths.append(f"{file_name}.yaml")

    return str(os.path.join(*os_paths))


async def generate_account_user_resource_files(
    exe_message: ExecutionMessage,
    aws_account: AWSAccount,
) -> dict:
    account_resource_dir = get_response_dir(exe_message, aws_account)
    user_resource_file_upsert_semaphore = NoqSemaphore(resource_file_upsert, 10)
    messages = []

    response = dict(account_id=aws_account.account_id, users=[])
    iam_client = await aws_account.get_boto3_client("iam")
    account_users = await list_users(iam_client)

    log.debug(
        "Retrieved AWS IAM Users.",
        account_id=aws_account.account_id,
        account_name=aws_account.account_name,
        user_count=len(account_users),
    )

    for account_user in account_users:
        user_path = os.path.join(
            account_resource_dir, f'{account_user["UserName"]}.json'
        )
        response["users"].append(
            {
                "path": user_path,
                "name": account_user["UserName"],
                "account_id": aws_account.account_id,
            }
        )
        messages.append(
            dict(file_path=user_path, content_as_dict=account_user, replace_file=True)
        )

    await user_resource_file_upsert_semaphore.process(messages)
    log.debug(
        "Finished caching AWS IAM Users.",
        account_id=aws_account.account_id,
        user_count=len(account_users),
    )

    return response


async def generate_user_resource_file_for_all_accounts(
    exe_message: ExecutionMessage,
    user_name: str,
    aws_account_map: dict[str, AWSAccount],
    updated_account_ids: list[str],
    iambic_template: Optional[AwsIamUserTemplate],
) -> list:
    async def get_user_for_account(aws_account: AWSAccount):
        iam_client = await aws_account.get_boto3_client("iam")
        return {aws_account.account_id: await get_user(user_name, iam_client)}

    account_resource_dir_map = {
        aws_account.account_id: get_response_dir(exe_message, aws_account)
        for aws_account in aws_account_map.values()
    }
    user_resource_file_upsert_semaphore = NoqSemaphore(resource_file_upsert, 10)
    messages = []
    response = []

    user_across_accounts = generate_template_output(
        updated_account_ids, aws_account_map, iambic_template
    )
    updated_account_users = await asyncio.gather(
        *[
            get_user_for_account(aws_account_map[account_id])
            for account_id in updated_account_ids
        ]
    )
    for updated_account_user in updated_account_users:
        user_across_accounts.update(updated_account_user)
    user_across_accounts = {k: v for k, v in user_across_accounts.items() if v}

    log.debug(
        "Retrieved AWS IAM User for all accounts.",
        user_name=user_name,
        total_accounts=len(user_across_accounts),
    )

    for account_id, account_user in user_across_accounts.items():
        user_path = os.path.join(
            account_resource_dir_map[account_id],
            f'{account_user["UserName"]}.json',
        )
        response.append(
            {
                "path": user_path,
                "name": account_user["UserName"],
                "account_id": account_id,
            }
        )
        messages.append(
            dict(file_path=user_path, content_as_dict=account_user, replace_file=True)
        )

    await user_resource_file_upsert_semaphore.process(messages)
    log.debug(
        "Finished caching AWS IAM User for all accounts.",
        user_name=user_name,
        total_accounts=len(user_across_accounts),
    )

    return response


async def set_user_resource_tags(
    user_name: str, user_resource_path: str, aws_account: AWSAccount
):
    iam_client = await aws_account.get_boto3_client("iam")
    user_tags = await list_user_tags(user_name, iam_client)
    await resource_file_upsert(user_resource_path, {"Tags": user_tags}, False)


async def set_user_resource_inline_policies(
    user_name: str, user_resource_path: str, aws_account: AWSAccount
):
    iam_client = await aws_account.get_boto3_client("iam")
    user_inline_policies = await get_user_inline_policies(user_name, iam_client)
    for k in user_inline_policies.keys():
        user_inline_policies[k]["policy_name"] = k

    user_inline_policies = list(user_inline_policies.values())
    await resource_file_upsert(
        user_resource_path, {"InlinePolicies": user_inline_policies}, False
    )


async def set_user_resource_managed_policies(
    user_name: str, user_resource_path: str, aws_account: AWSAccount
):
    iam_client = await aws_account.get_boto3_client("iam")
    user_managed_policies = await get_user_managed_policies(user_name, iam_client)
    await resource_file_upsert(
        user_resource_path, {"ManagedPolicies": user_managed_policies}, False
    )


async def set_user_resource_groups(
    user_name: str, user_resource_path: str, aws_account: AWSAccount
):
    iam_client = await aws_account.get_boto3_client("iam")
    user_groups = await get_user_groups(user_name, iam_client)
    await resource_file_upsert(user_resource_path, {"Groups": user_groups}, False)


async def _account_id_to_user_map(user_refs):
    account_id_to_user_map = {}
    for user_ref in user_refs:
        async with aiofiles.open(user_ref["path"], mode="r") as f:
            content_dict = json.loads(await f.read())
            account_id_to_user_map[user_ref["account_id"]] = normalize_dict_keys(
                content_dict
            )
    return account_id_to_user_map


async def create_templated_user(  # noqa: C901
    aws_account_map: dict[str, AWSAccount],
    user_name: str,
    user_refs: list[dict],
    user_dir: str,
    existing_template_map: dict,
    config: AWSConfig,
) -> Optional[AwsIamUserTemplate]:
    account_id_to_user_map = await _account_id_to_user_map(user_refs)
    num_of_accounts = len(user_refs)

    min_accounts_required_for_wildcard_included_accounts = (
        config.min_accounts_required_for_wildcard_included_accounts
    )

    # calculate preference based on existing template
    prefer_templatized = calculate_import_preference(
        existing_template_map.get(user_name)
    )

    # Generate the params used for attribute creation
    user_template_params = {"identifier": user_name}
    user_template_properties = {"user_name": user_name}
    path_resources = list()
    description_resources = list()
    managed_policy_resources = list()
    inline_policy_document_resources = list()
    permissions_boundary_resources = list()
    group_resources = list()
    tag_resources = list()
    import_actions = set()

    for account_id, user_dict in account_id_to_user_map.items():
        import_actions.update(
            await process_import_rules(
                config,
                AWS_IAM_USER_TEMPLATE_TYPE,
                user_name,
                user_dict.get("tags", []),
                user_dict,
            )
        )
        path_resources.append(
            {
                "account_id": account_id,
                "resources": [{"resource_val": user_dict["path"]}],
            }
        )

        if managed_policies := user_dict.get("managed_policies"):
            managed_policy_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": mp} for mp in managed_policies],
                }
            )

        if groups := user_dict.get("groups"):
            group_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": v} for _, v in groups.items()],
                }
            )

        if permissions_boundary := user_dict.get("permissions_boundary"):
            permissions_boundary_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": permissions_boundary}],
                }
            )

        if inline_policies := user_dict.get("inline_policies"):
            # Normalize the inline policy statements to be a list
            for inline_policy in inline_policies:
                if isinstance(inline_policy.get("statement"), dict):
                    inline_policy["statement"] = [inline_policy["statement"]]
            inline_policy_document_resources.append(
                {
                    "account_id": account_id,
                    "resources": [
                        {"resource_val": inline_policy}
                        for inline_policy in inline_policies
                    ],
                }
            )

        if tags := user_dict.get("tags"):
            tag_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": tag} for tag in tags],
                }
            )

        if description := user_dict.get("description"):
            description_resources.append(
                {"account_id": account_id, "resources": [{"resource_val": description}]}
            )
    for action in import_actions:
        if action.value == "set_import_only":
            user_template_params["iambic_managed"] = "import_only"
        if action.value == "ignore":
            return None
    if (
        len(user_refs) != len(aws_account_map)
        or len(aws_account_map) <= min_accounts_required_for_wildcard_included_accounts
    ):
        user_template_params["included_accounts"] = [
            aws_account_map[user_ref["account_id"]].account_name
            for user_ref in user_refs
        ]
    else:
        user_template_params["included_accounts"] = ["*"]

    path = await group_int_or_str_attribute(
        aws_account_map, num_of_accounts, path_resources, "path"
    )
    if path != "/":
        user_template_properties["path"] = path

    if permissions_boundary_resources:
        user_template_properties["permissions_boundary"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            permissions_boundary_resources,
            prefer_templatized=prefer_templatized,
        )

    if description_resources:
        user_template_properties["description"] = await group_int_or_str_attribute(
            aws_account_map, num_of_accounts, description_resources, "description"
        )

    if managed_policy_resources:
        user_template_properties["managed_policies"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            managed_policy_resources,
            False,
            prefer_templatized=prefer_templatized,
        )

    if inline_policy_document_resources:
        user_template_properties["inline_policies"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            inline_policy_document_resources,
            False,
            prefer_templatized=prefer_templatized,
        )

    if group_resources:
        user_template_properties["groups"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            group_resources,
            False,
            prefer_templatized=prefer_templatized,
        )
    if tag_resources:
        tags = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            tag_resources,
            True,
            prefer_templatized=prefer_templatized,
        )
        if isinstance(tags, dict):
            tags = [tags]

        user_template_properties["tags"] = tags

    file_path = get_templated_user_file_path(
        user_dir,
        user_name,
        user_template_params.get("included_accounts"),
        user_path=path,
    )
    return create_or_update_template(
        file_path,
        existing_template_map,
        user_name,
        AwsIamUserTemplate,
        user_template_params,
        UserProperties(**user_template_properties),
        list(aws_account_map.values()),
    )


async def collect_aws_users(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    iam_template_map: dict,
    detect_messages: list[UserMessageDetails] = None,
):
    aws_account_map = await get_aws_account_map(config)
    if exe_message.provider_id:
        aws_account_map = {
            exe_message.provider_id: aws_account_map[exe_message.provider_id]
        }

    existing_template_map = iam_template_map.get(AWS_IAM_USER_TEMPLATE_TYPE, {})
    set_user_resource_inline_policies_semaphore = NoqSemaphore(
        set_user_resource_inline_policies, 20
    )
    set_user_resource_managed_policies_semaphore = NoqSemaphore(
        set_user_resource_managed_policies, 30
    )

    set_user_resource_groups_semaphore = NoqSemaphore(set_user_resource_groups, 25)
    set_user_resource_tags_semaphore = NoqSemaphore(set_user_resource_tags, 30)

    log.info(
        "Generating AWS user templates. Beginning to retrieve AWS IAM Users.",
        accounts=list(aws_account_map.keys()),
    )

    if detect_messages:
        generate_user_resource_file_for_all_accounts_semaphore = NoqSemaphore(
            generate_user_resource_file_for_all_accounts, 50
        )
        grouped_detect_messages = group_detect_messages("user_name", detect_messages)
        tasks = [
            {
                "exe_message": exe_message,
                "aws_account_map": aws_account_map,
                "user_name": user_name,
                "updated_account_ids": [
                    message.account_id for message in user_messages
                ],
                "iambic_template": existing_template_map.get(user_name),
            }
            for user_name, user_messages in grouped_detect_messages.items()
        ]

        # Remove deleted or mark templates for update
        deleted_users = [user for user in detect_messages if user.delete]
        if deleted_users:
            for user in deleted_users:
                user_account = aws_account_map.get(user.account_id)
                if existing_template := existing_template_map.get(user.user_name):
                    if len(existing_template.included_accounts) == 1 and (
                        existing_template.included_accounts[0]
                        == user_account.account_name
                        or existing_template.included_accounts[0]
                        == user_account.account_id
                    ):
                        # It's the only account for the template so delete it
                        existing_template.delete()

        account_user_list = (
            await generate_user_resource_file_for_all_accounts_semaphore.process(tasks)
        )
        account_user_list = list(itertools.chain.from_iterable(account_user_list))
        account_user_map = defaultdict(list)
        for account_user in account_user_list:
            account_user_map[account_user["account_id"]].append(account_user)
        account_users = [
            dict(account_id=account_id, users=users)
            for account_id, users in account_user_map.items()
        ]
    elif exe_message.provider_id:
        aws_account = aws_account_map[exe_message.provider_id]
        account_users = [
            (await generate_account_user_resource_files(exe_message, aws_account))
        ]
    else:
        generate_account_user_resource_files_semaphore = NoqSemaphore(
            generate_account_user_resource_files, 5
        )
        account_users = await generate_account_user_resource_files_semaphore.process(
            [
                {"exe_message": exe_message, "aws_account": aws_account}
                for aws_account in aws_account_map.values()
            ]
        )

    messages = []
    # Upsert users
    for account_user in account_users:
        for user in account_user["users"]:
            messages.append(
                {
                    "user_name": user["name"],
                    "user_resource_path": user["path"],
                    "aws_account": aws_account_map[account_user["account_id"]],
                }
            )

    if not detect_messages:
        log.info(
            "Setting inline policies in user templates",
            accounts=list(aws_account_map.keys()),
        )
        await set_user_resource_inline_policies_semaphore.process(messages)
        log.info(
            "Setting managed policies in user templates",
            accounts=list(aws_account_map.keys()),
        )
        await set_user_resource_managed_policies_semaphore.process(messages)
        log.info(
            "Setting groups in user templates", accounts=list(aws_account_map.keys())
        )
        await set_user_resource_groups_semaphore.process(messages)
        log.info(
            "Setting tags in user templates", accounts=list(aws_account_map.keys())
        )
        await set_user_resource_tags_semaphore.process(messages)
        log.info(
            "Finished retrieving user details", accounts=list(aws_account_map.keys())
        )

    account_user_output = json.dumps(account_users)
    with open(
        exe_message.get_file_path(*RESOURCE_DIR, file_name_and_extension="output.json"),
        "w",
    ) as f:
        f.write(account_user_output)


async def generate_aws_user_templates(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    base_output_dir: str,
    iam_template_map: dict,
    detect_messages: list[UserMessageDetails] = None,
):
    aws_account_map = await get_aws_account_map(config)

    if detect_messages:
        detect_messages = [
            msg for msg in detect_messages if isinstance(msg, UserMessageDetails)
        ]
        if not detect_messages:
            return

    existing_template_map = iam_template_map.get(AWS_IAM_USER_TEMPLATE_TYPE, {})
    user_dir = get_template_dir(base_output_dir)
    account_users = await exe_message.get_sub_exe_files(
        *RESOURCE_DIR, file_name_and_extension="output.json", flatten_results=True
    )

    log.info("Grouping users")
    # Move everything to required structure
    for account_user_elem in range(len(account_users)):
        for user_elem in range(len(account_users[account_user_elem]["users"])):
            user_name = account_users[account_user_elem]["users"][user_elem].pop("name")
            account_users[account_user_elem]["users"][user_elem][
                "resource_val"
            ] = user_name

        account_users[account_user_elem]["resources"] = account_users[
            account_user_elem
        ].pop("users", [])

    grouped_user_map = await base_group_str_attribute(aws_account_map, account_users)

    log.debug("Writing templated users")
    all_resource_ids = set()
    for user_name, user_refs in grouped_user_map.items():
        resource_template = await create_templated_user(
            aws_account_map,
            user_name,
            user_refs,
            user_dir,
            existing_template_map,
            config,
        )
        if not resource_template:
            # Template not updated. Most likely because it's an `enforced` template.
            continue
        all_resource_ids.add(resource_template.resource_id)

    if not detect_messages:
        # NEVER call this if messages are passed in because all_resource_ids will only contain those resources
        delete_orphaned_templates(
            list(existing_template_map.values()), all_resource_ids
        )

    log.info("Finished templated user generation")
